/** Änderungschronik
    Vers. 0.14
    27.03.2020: Das Retain-Flag wird beim MQTT-publishing gesetzt
    05.04.2020: Timeservermethoden werden verbessert und in eigene library ausgelagert
*/

#include <PubSubClient.h>
#include <ESP8266WiFi.h>
#include <OneWire.h>
#include <DallasTemperature.h>
#include <WiFiClient.h>
#include <ESP8266mDNS.h>
//#include <WiFiUdp.h>
#include <ESP8266WebServer.h>
#include <ESP8266HTTPUpdateServer.h>
#include <Timeserver.h>

//**** Diese Daten müssen angepasst werden ***************

#define WIFI_AP "YOUR_SSID"
#define WIFI_PASSWORD "YOUR_PASSWORD"

#define TOKEN "ESP8266_DALLAS_TOKEN_SZ"
#define ONE_WIRE_BUS 2  // DS18B20 pin
#define FAKTOR 60000  //  60.000ms = 1 min

// Adresse eines Timeservers - hier im internen Netz
// alternativ kann hier auch die IP eines externen ntp Servers stehen
IPAddress timeServerIP(192, 168, xxx, xxx); 
Timeserver timeserver(timeServerIP);

String vers = "0.15";
String versionDate = "05.04.2020";
int interval = 5;
char clientID[] = "DALLAS_Sensor_Sz";
char mqttBroker[] = "192.168.xxx.yyy"; // Adresse des MQTT-Brokers
int subNetIP = xxx; //IP des Subnets
int myIP = zzz; //nur der letzte Block der IP ändert sich
String myHostname = "Schlafzimmer";
String mySensor = "DS18B20";
String mqttSubject = "wetter/sz/temperatur";
String temperatur = "no Data";
unsigned long lastMessage;


//**********************************************************

/* Variablen und Konstanten für die Timeserver-Methode
  unsigned int localPort = 2390; //local UDP - Port
  IPAddress timeServerIP; // time.nist.gov NTP server address
  const char* ntpServerName = "time.nist.gov";
  const int NTP_PACKET_SIZE = 48; // NTP time stamp is in the first 48 bytes of the message
  byte packetBuffer[ NTP_PACKET_SIZE]; //buffer to hold incoming and outgoing packets
  // A UDP instance to let us send and receive packets over UDP
  WiFiUDP udp;
  // Ende Time-Server Geraffel */

// DALLAS

OneWire oneWire(ONE_WIRE_BUS);
DallasTemperature DS18B20(&oneWire);

WiFiClient wifiClient;

PubSubClient client(wifiClient);

ESP8266WebServer server(80); //Server on port 80
ESP8266HTTPUpdateServer httpUpdater;

int status = WL_IDLE_STATUS;

void setup()
{
  //Serial.begin(115200);
  InitWiFi();
  server.on("/", handleRoot); //Routine to handle at root location
  server.on("/a1", []() {
    interval = 1;
    gotoRoot("1");
  });
  server.on("/b2", []() {
    interval = 2;
    gotoRoot("2");
  });
  server.on("/c5", []() {
    interval = 5;
    gotoRoot("5");
  });
  server.on("/d10", []() {
    interval = 10;
    gotoRoot("10");
  });
  server.on("/e15", []() {
    interval = 15;
    gotoRoot("15");
  });
  server.on("/f30", []() {
    interval = 30;
    gotoRoot("30");
  });
  httpUpdater.setup(&server);
  server.begin();
  client.setServer( mqttBroker, 1883 );
  lastMessage = millis() + (interval * FAKTOR);
}

void loop() {
  if ( !client.connected() ) {
    reconnect();
  }
  server.handleClient();
  if ( millis() - lastMessage > interval * FAKTOR ) {
    temperatur = getData();
    sendMQTTMessage(mqttSubject, temperatur);
    lastMessage = millis();
  }
  client.loop();
}

String getFormattedFloat(float x, uint8_t precision) {
  char buffer[10];
  dtostrf(x, 7, precision, buffer);
  return buffer;
}

String getData() {
  float temp;
  do {
    DS18B20.requestTemperatures();
    //Serial.println("T requested");
    temp = DS18B20.getTempCByIndex(0);
  } while (temp == 85.0 || temp == (-127.0));
  // Serial.println("got DATA");
  String temperature = getFormattedFloat(temp, 1);
  //Serial.print("formatted Temperature: ");
  //Serial.println(temperature);
  return temperature;
}

void sendMQTTMessage(String subject, String message) { 
  char payload[100];
  char sub[100];
  message.toCharArray( payload, 100 );
  subject.toCharArray(sub, 100);
  //Serial.print("MQTT-Payload: ");
  //Serial.println(message);
  //Serial.print("MQTT-Subject: ");
  //Serial.println(sub);
  client.publish( sub, payload , true);
}

void connectToWiFi() {
  WiFi.begin(WIFI_AP, WIFI_PASSWORD);
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    // Serial.print(".");
  }
  //Serial.println("Connected to AccessPoint");
}


String macAdresse() {
  String out = "";
  byte mac[6];
  WiFi.macAddress(mac);
  for (int i = 5; i > 0; i--) {
    out += hex(mac[i]) + ":";
  }
  out += hex(mac[0]);
  return out;
}

String hex(byte b) {
  String out = String(b, HEX);
  if (out.length() == 1) out = "0" + out;
  return out;
}

void InitWiFi()
{
  //Serial.println("Connecting to AccessPoint ...");
  // attempt to connect to WiFi network
  IPAddress ip(192, 168, subNetIP, myIP);
  IPAddress gateway(192, 168, subNetIP, 1);
  IPAddress subnet(255, 255, 255, 0);
  IPAddress dns(192, 168, subNetIP, 1);
  WiFi.config(ip, dns, gateway, subnet); // auf feste IP einstellen
  WiFi.hostname(myHostname); //Hostnamen setzen
  WiFi.persistent(false);
  WiFi.mode(WIFI_OFF);
  WiFi.mode(WIFI_STA);
  connectToWiFi();
}

void reconnect() {
  // Loop until we're reconnected
  while (!client.connected()) {
    status = WiFi.status();
    if ( status != WL_CONNECTED) {
      connectToWiFi();
    }
    //Serial.print("Connecting to MQTT_Broker ...");
    // Attempt to connect (clientId, username, password)
    if ( client.connect(clientID, TOKEN, NULL) ) {
      //Serial.println( "[DONE]" );
    } else {
      //Serial.print( "[FAILED] [ rc = " );
      //Serial.print( client.state() );
      //Serial.println( " : retrying in 5 seconds]" );
      // Wait 5 seconds before retrying
      delay( 5000 );
    }
  }
}

String inttostr(int i) {
  char cstring[4];
  sprintf(cstring, "%d", i);
  return (String) cstring;
}

String messString(String data, String art, String einheit) {
  String out = art + ": ";
  if (data == "no data") out += "Ablesefeher!";
  else {
    out += data + einheit;
  }
  return out;
}

String onlineTime() {
  unsigned long t = millis();
  String out = "";
  t = t / 1000; // Onlinezeit in Sekunden
  out += String((int)(t / 86400L)) + " d, ";
  t = t % 86400L;
  out += String((int)(t / 3600L)) + " h, ";
  t = t % 3600L;
  out += String((int)(t / 60)) + " m, " + String((int)(t % 60)) + " s";
  return out;
}

void handleRoot() {
  String msg = "<meta http-equiv=\"refresh\" content=\"60\"> ";
  msg += "<center><H1><u>" + String(myHostname) + "</u></H1> </p>";
  msg += "<H3> Sensor: " + String(mySensor) + " (Temperatur)</H3></p>";
  msg += "Online seit: " + onlineTime() + "<br />";
  msg += "<H2><u>" + messString(temperatur, "Temperatur", " &deg;C") + "</u></H2> </p>";
  msg += "<H3>Abgelesen am: " + timeserver.dateTimeString(timeserver.UNIT) + "</H3></p>";
  msg += "<b>MQTT-Token:</b><i> " + String(mqttSubject) + "</i><br />";
  msg += "<b>IP-Adresse:</b> 192.168.12." + inttostr(myIP) + "<br />";
  msg += "<b>MAC-Adresse:</b> " + macAdresse() + "<br />";
  msg += "<H3>Messintervall: " + inttostr(interval) + " min</H3></p>";
  msg += "===================================================<br />";
  msg += "Messintervall ausw&auml;hlen:<br>";
  msg += "<a href = \"a1\">eine Minute</a><br />\n";
  msg += "<a href = \"b2\">zwei Minuten</a><br />\n";
  msg += "<a href = \"c5\">f&uuml;nf Minuten</a><br />\n";
  msg += "<a href = \"d10\">zehn Minuten</a><br />\n";
  msg += "<a href = \"e15\">f&uuml;nfzehn Minuten</a><br />\n";
  msg += "<a href = \"f30\">dreissig Minuten</a><br />\n";
  msg += "===================================================<br />";
  msg += "Firmware Version " + vers + " vom " + versionDate + "<br />";
  msg += "<H2><a href=\"update\">Firmware-Update</a></H2><br /></center>";
  server.send(200, "text/html", msg);
}

void gotoRoot(String intervall) {
  String msg = "<html><head><title>Intervall gesetzt</title></head><body>";
  msg += "<center><br><br><br><br><br><H1>Das Messintervall betr&auml;gt nun " + intervall + " Minuten</H1>";
  msg += "<a href = \"/\"> zur&uuml;ck zur Hauptseite</a></center></body></html>";
  server.send(200, "text/html", msg);
}
