/**
 Timeserver Library
 Ideen dazu von https://circuits4you.com 
 Erweitert durch Reiner Hornischer
 http://www.hornischer.net
*/

#include "Timeserver.h"
#include <ESP8266WiFi.h>

Timeserver::Timeserver (IPAddress timeServerIP){
		this->timeServerIP = timeServerIP;
		udp.begin(localPort);
	}
	
/*Timeserver::Timeserver (String ntpServerName){
		WiFi.hostByName(ntpServerName, this->timeServerIP); 
		//this->timeServerIP = timeServerIP;
		udp.begin(localPort);
	}*/

/*Timeserver::Timeserver(){
	}

void Timeserver::connect(String ntpServerName){
		WiFi.hostByName(ntpServerName, this->timeServerIP); 
		udp.begin(localPort);
	}*/

unsigned long serverTime = 0;
int deltadays = 0;
int day = 0;
int dow=0;
int month = 0;
int year = 0;
int hour = 0;
int minute = 0;
int second = 0;

String dayname[]={"Sonntag","Montag","Dienstag","Mittwoch","Donnerstag","Freitag","Samstag"};
	
String Timeserver::niceTime(int t){
	String out = "";
	if (t<10) out += "0";
	out+=String(t);
	return out;
}

String Timeserver::timeString(bool raw, bool recalc) {
		if (recalc) calcTime();
		String sts=" Uhr ";
		if (isSummerTime()) sts+="MESZ"; else sts+="MEZ"; 
		if (raw) return String(hour)+":"+niceTime(minute)+":"+niceTime(second);
		return String(hour)+":"+niceTime(minute)+":"+niceTime(second)+sts;
	}
	
String Timeserver::dateString(bool recalc){
		if (recalc) calcTime();
		return dayname[dow]+", "+niceTime(day)+"."+niceTime(month)+"."+year;
	}
	
String Timeserver::dateTimeString(bool raw) {	
		String out = dateString(true)+" - ";
		out+=timeString(raw, false);
		return out;
	}
	
unsigned long Timeserver::dateTime(){
		unsigned long out = ((((year * 10 + month)*10 + day)*10 + hour)*10 + minute)*10 + second;
		return out;
	}	
	
void Timeserver::calcTime(){
		serverTime = requestTime();
		calcyear();
		calcmonth();
		calcday();
		dow=calcDoW(day,month,year);
		hour = ((serverTime % 86400L) / 3600)+1;
		if (isSummerTime()) hour++;
		minute = (serverTime %3600) /60;
		second = serverTime %60;
	}

int Timeserver::getYear(){
		return year;
	}
	
int Timeserver::getMonth(){
		return month;
	}

int Timeserver::getDay(){
		return day;
	}
	
int Timeserver::getHour(){
		return hour;
	}
	
int Timeserver::getMinute(){
		return minute;
	}
	
int Timeserver::getSecond(){
		return second;
	}

int tageAbStart[] = {31,59,90,120,151,181,212,243,273,304,334,365};

void Timeserver::calcyear(){
		int days = serverTime/86400;
		int fourYearBlocks = days/1461;
		int years=fourYearBlocks*4;
		days = days % 1461;
		years += days / 365;
		deltadays=days%365;
		year=years +2000;
	}
bool Timeserver::isSchaltjahr(){
		return (year %4 == 0);
	}
	
void Timeserver::calcmonth(){
		if (deltadays < 31) month = 1;
		if (isSchaltjahr()) deltadays-=1;
		int i=0;
		do {
			i++;
		}
		while (deltadays>tageAbStart[i]);
		month = ++i;
	}
	
void Timeserver::calcday(){
		day = deltadays-tageAbStart[month-2];
		if (isSchaltjahr()) day++;
	}

int Timeserver::calcDoW(int d, int m, int y){
		int nm[] = {0,3,3,6,1,4,6,2,5,0,3,5};
		int nt=d%7;
		int jjh=y-2000;
		int jz=(jjh+(jjh/4))%7;
		int jh=6; //gilt nur im Jahr 20xx
		int sj=0;
		if (isSchaltjahr() && (m<3)) sj=6;
		return((nt+jz+jh+sj+nm[--m])%7);
	}

bool Timeserver::isSummerTime(){
		int d31 = calcDoW(31,3,year);
		int summertimeday=31-d31;
		d31 = calcDoW(31,10,year);
		int wintertimeday=31-d31;
		return (((month==3 && day>=summertimeday&&hour>1) || month>3) && ((month==10 && day<wintertimeday && hour<2)||month<10));
	}

// Gibt die Zeit seit 1970 in Sekunden zurück
unsigned long Timeserver::requestTime() {
		bool received = false;
		unsigned long out = 0;
		while (!received) {
			//WiFi.hostByName(ntpServerName, timeServerIP);
			sendNTPpacket(timeServerIP); // send an NTP packet to a time server
			// wait to see if a reply is available
			delay(1000);
			int cb = udp.parsePacket();
			if (cb) {
			received = true;
			udp.read(packetBuffer, Timeserver::NTP_PACKET_SIZE); // read the packet into the buffer
			unsigned long highWord = word(packetBuffer[40], packetBuffer[41]);
			unsigned long lowWord = word(packetBuffer[42], packetBuffer[43]);
			// combine the four bytes (two words) into a long integer
			// this is NTP time (seconds since Jan 1 1900):
			unsigned long secsSince1900 = highWord << 16 | lowWord;
			// now convert NTP time into everyday time:
			//Serial.print("Unix time = ");
			// Unix time starts on Jan 1 1970. In seconds, that's 2208988800:
			//const unsigned long seventyYears = 2208988800UL;
			const unsigned long hundredYears = 3155673600UL;
			// subtract seventy years:
			//out = secsSince1900 - seventyYears;
			out = secsSince1900 - hundredYears;
			} else  delay(10000);
		}
		return out;
	}
	
	// send an NTP request to the time server at the given address
unsigned long Timeserver::sendNTPpacket(IPAddress& address)
	{
		//Serial.println("sending NTP packet...");
		// set all bytes in the buffer to 0
		memset(packetBuffer, 0, NTP_PACKET_SIZE);
		// Initialize values needed to form NTP request
		// (see URL above for details on the packets)
		packetBuffer[0] = 0b11100011;   // LI, Version, Mode
		packetBuffer[1] = 0;     // Stratum, or type of clock
		packetBuffer[2] = 6;     // Polling Interval
		packetBuffer[3] = 0xEC;  // Peer Clock Precision
		// 8 bytes of zero for Root Delay & Root Dispersion
		packetBuffer[12]  = 49;
		packetBuffer[13]  = 0x4E;
		packetBuffer[14]  = 49;
		packetBuffer[15]  = 52;

		// all NTP fields have been given values, now
		// you can send a packet requesting a timestamp:
		udp.beginPacket(address, 123); //NTP requests are to port 123
		udp.write(packetBuffer, NTP_PACKET_SIZE);
		udp.endPacket();
	}
